/*-----------------------------------------------------------------------
cdgprmacros.sas

SAS program called by cdgprmainprogram.sas, part of the replication files for
'The productivity advantages of large cities: Distinguishing agglomeration from firm selection'
by Pierre-Philippe Combes, Gilles Duranton, Laurent Gobillon, Diego Puga, and Sbastien Roux. 
-----------------------------------------------------------------------*/

/* macro that specifies the parametric functions of value */
%macro specif(truncation,shift,dilation);
	/* value transformation function */
	start zb(u,beta);

	%if &shift=0 %then
		%do;
			zb=beta*u;
		%end;
	%else %if &dilation=0 %then
		%do;
			zb=u+beta;
		%end;
	%else
		%do;
			zb=beta[1]*u+beta[2];
		%end;

	return(zb);
	finish;

	/* partial derivative of zb wrt beta */
	start dzb_b(u,beta);

	%if &shift=0 %then
		%do;
			dzb=u;
		%end;
	%else %if &dilation=0 %then
		%do;
			dzb=shape(1,nrow(u),1);
		%end;
	%else
		%do;
			dzb=u||shape(1,nrow(u),1);
		%end;

	return(dzb);
	finish;

	/* partial derivative of zb wrt u */
	start dzb_u(u,beta);

	%if &shift=0 %then
		%do;
			dzb=beta[1];
		%end;
	%else %if &dilation=0 %then
		%do;
			dzb=1;
		%end;
	%else
		%do;
			dzb=beta[1];
		%end;

	return(dzb);
	finish;

	/* inverse of zb, holding beta constant */
	start zb_1(x,beta);

	%if &shift=0 %then
		%do;
			zb_1=x/beta;
		%end;
	%else %if &dilation=0 %then
		%do;
			zb_1=x-beta;
		%end;
	%else
		%do;
			zb_1=(x-beta[2])/beta[1];
		%end;

	return(zb_1);
	finish;

	/* derivative of zb_1 wrt beta */
	start dzb_1_b(x,beta);

	%if &shift=0 %then
		%do;
			dzb=-x/(beta**2);
		%end;
	%else %if &dilation=0 %then
		%do;
			dzb=-shape(1,nrow(x),1);
		%end;
	%else
		%do;
			dzb=(-(x-beta[2])/(beta[1]**2))||(-1/beta[1]*shape(1,nrow(x),1));
		%end;

	return(dzb);
	finish;
%mend;

/* specification of the function of rank transformation, or selection */
%macro specifgamm;
	start gammap(pgamm);
	gamm=(exp(pgamm[1])-5)/(5+exp(pgamm[1]));
	return(gamm);
	finish;
	start gamma_1(gamm);
	pgamm=log((1+gamm)/(1-gamm))+log(5);
	return(pgamm);
	finish;
	start tg(u,gamma);
	tg=gamma+(1-gamma)*u;
	return(tg);
	finish;
	start dtg_g(u,gamma);
	dtg=(1-u);
	return(dtg);
	finish;
	start dtg_u(u,gamma);
	dtg=shape((1-gamma),nrow(u),1);
	return(dtg);
	finish;
	start tg_1(x,gamma);
	tg_1=(x-gamma)/(1-gamma);
	return(tg_1);
	finish;
%mend;

/* macro optimizing the criteria wrt to beta (and not gamma) */
%macro init(truncation,shift,dilation);
	%if &shift=0 & &dilation=0 %then
		%do;
			betaf=1;
		%end;
	%else
		%do;
			/* initialisation of the parameters using values computed in the macro crit */
			%if &shift=0 %then
				%do;
					beta0=sdyn;
				%end;
			%else %if &dilation=0 %then
				%do;
					beta0=myn2-myn1;
				%end;
			%else
				%do;
					beta0=sdyn//(myn2-myn1*sdyn);
				%end;

			opt={0 0};

			/* optimization routines with SAS-IML, critb is the function to minimize, and 
				dcritb is derivative */
			tcr = j(13,1,.);
			tcr[4] = 1e-10;
			call nlpqn(rc,betaf,"critb",beta0,,,tcr,,,"dcritb",,);

			if rc<0 then
				do;
					call nlpnms(rc,betaf,"critb",beta0,,,r,,,);
				end;
		%end;
%mend;

/* macro estimating parameters for the considered transformation */
%macro estquant(variable,data,varcat,varref0,varref1,truncation,shift,dilation);

	proc iml;
		%specif(&truncation,&shift,&dilation);
		%specifgamm;

		/*allowing for weights, here uniform*/
		start weight1(x);
			x=shape(1,nrow(x),ncol(x));
			return(x);
		finish;
		start weight2(x);
			x=shape(1,nrow(x),ncol(x));
			return(x);
		finish;

		/* procedure estimating quantiles of rank rang2 from the distribution of values in y1
		using linear interpolation */
		start quant(y1,rang2);
			nobs1=nrow(y1);
			intb=int(nobs1*rang2);
			nobs2=nrow(rang2);
			intb=intb+(intb=J(nobs2,1,0));
			intu=int(nobs1*rang2)+1;
			intu=intu-(intu>J(nobs2,1,nobs1));
			qy1b=y1[intb];
			qy1u=y1[intu];
			qy1=(rang2*nobs1-intb)#(qy1u-qy1b)+qy1b;
			return(qy1);
		finish;

		/* computation of the criterium to minimize, as a function of beta only */
		start critb(beta) global(y,Ct1,Ct2);
			posb=(1:(nrow(y)-1));
			posu=(2:nrow(y));
			call sort(y,3);
			w1=y[,5];
			yt1=y[,1];
			yt2=y[,2];
			rgt1=y[,3];
			It1=w1#(zb(yt1,beta)-yt2)##2;
			Ct1=t(It1[posu]+It1[posb])*(rgt1[posu]-rgt1[posb])/(rgt1[nrow(y)]-rgt1[1])/2;
			call sort(y,4);
			w2=y[,6];
			yt1=y[,1];
			yt2=y[,2];
			rgt2=y[,4];
			It2=w2#(yt1-zb_1(yt2,beta))##2;
			Ct2=t(It2[posu]+It2[posb])*(rgt2[posu]-rgt2[posb])/(rgt2[nrow(y)]-rgt2[1])/2;
			return(Ct1+Ct2);
		finish;

		/* computation of the derivative of the criterium, wrt beta */
		start dcritb(beta) global(y);
			posb=(1:(nrow(y)-1));
			posu=(2:nrow(y));
			call sort(y,3);
			w1=y[,5];
			yt1=y[,1];
			yt2=y[,2];
			rgt1=y[,3];
			It1=2*(shape(1,1,nrow(beta))@(w1#(zb(yt1,beta)-yt2)))#dzb_b(yt1,beta);
			Ct1=t(It1[posu,]+It1[posb,])*(rgt1[posu]-rgt1[posb])/(rgt1[nrow(y)]-rgt1[1])/2;
			call sort(y,4);
			w2=y[,6];
			yt1=y[,1];
			yt2=y[,2];
			rgt2=y[,4];
			It2=-2*(shape(1,1,nrow(beta))@(w2#(yt1-zb_1(yt2,beta))))#dzb_1_b(yt2,beta);
			Ct2=t(It2[posu,]+It2[posb,])*(rgt2[posu]-rgt2[posb])/(rgt2[nrow(y)]-rgt2[1])/2;
			return(t(Ct1+Ct2));
		finish;

		/* non-parametric computation of quantile derivatives, used for the test statistics */
		start BlGart(p,yv);
			np=nrow(p);
			n=nrow(yv);
			m=int(sqrt(n));
			vk1=(int(n*p)+m)#(n*p<=J(np,1,n-m))+n*(n*p>J(np,1,n-m));
			vk0=(int(n*p)-m)#(n*p>=J(np,1,1+m))+1*(n*p<J(np,1,1+m));
			lambdp=(yv[vk1]-yv[vk0])*n/(vk1-vk0);
			return(lambdp);
		finish;

		/* computation of the criterium as a function of pgamm, which represents the truncation. 
		Note that the criterium depends on two parameters: beta and pgamm. 
		Its minimization is done in two steps: 
		1. Holding pgamm constant, computation of beta that minimizes the criterium, cf macro init;
		the obtained criterium is just a function of pgamm.
		2. pgamm minimizes the function obtained in step 1; it means that for every iteration of 
		the minimization algorithm, new values of beta are estimated
		*/
		start critbisf(pgamm) global(y,y1,y2,alphmin,alphmax,betaf,n1t,n2t,nrang);
			gamm=gammap(pgamm);

			/* common support condition */
			umin1=max(0,tg_1(0,gamm));
			umax1=min(1,tg_1(1,gamm));
			umin2=max(0,tg(0,gamm));
			umax2=min(1,tg(1,gamm));

			/* rank transformation */
			n1t=int(nrow(y1)*(umax1-umin1));
			n2t=int(nrow(y2)*(umax2-umin2));

			/* computation of threshold ranks */
			rangn1=umin1+t(0:nrang)/nrang*(umax1-umin1);
			tgrangn1=tg(rangn1,gamm);
			rangn2=umin2+t(0:nrang)/nrang*(umax2-umin2);
			tg_rangn2=tg_1(rangn2,gamm);
			w1=weight1(rangn1//tg_rangn2);
			w2=weight2(tgrangn1//rangn2);
			yn1=quant(y1,rangn1);
			qy2=quant(y2,tgrangn1);
			yn2=quant(y2,rangn2);
			qy1=quant(y1,tg_rangn2);
			y=((yn1||qy2||rangn1||tgrangn1)//(qy1||yn2||tg_rangn2||rangn2))
			||w1||w2||(shape(1,nrow(yn1),1)//shape(2,nrow(yn2),1));
			myn1=yn1[:];
			myn2=yn2[:];

			/* mqy1=qy1[:];print mqy1 myn2;
				 mqy2=qy2[:];print mqy2 myn1; */
			vyn1=((yn1-myn1)##2)[:];
			vyn2=((yn2-myn2)##2)[:];
			sdyn=1;

			if vyn1>0 then
				sdyn=sqrt(vyn2/vyn1);

			%init(&truncation,&shift,&dilation);
			cr=critb(betaf);

			* print cr betaf;
			return(cr);
		finish;
		use &data.(where=(&varcat=&varref0 & &variable. ne .));
		read all var{"&variable"} into Y1;
		use &data.(where=(&varcat=&varref1 & &variable. ne .));
		read all var{"&variable"} into Y2;
		call sort(y1,1);
		call sort(y2,1);

		/* %put(&truncation &shift &dilation); */
		n1t=nrow(y1);
		n2t=nrow(y2);

		/* defines the size of the vector used to compute the criterium */
		nrang=1000;

		%if &truncation=1 %then
			%do;
				pgammi=-2;
				pgamm0=pgammi;
				voirs=critbisf(pgammi);

				/* grid search on pgamm to have a better initial value, before the optimization routine */
				*print n1t n2t;
				table=t(pgammi)||voirs||n1t||n2t;

				do c=1 to 10;
					pgammi=pgammi+0.5;
					voiri=critbisf(pgammi);

					if voiri<voirs then
						do;
							pgamm0=pgammi;
							voirs=voiri;
						end;

					table=table//(t(gammap(pgammi))||voiri||n1t||n2t);
				end;

				pgamm=pgamm0;

				*print table;
				cr=critbisf(pgamm);
				opt={0 4};
				tcr = j(13,1,.);
				tcr[4] = 1e-10;
				call nlpqn(rc,pgammf,"critbisf",pgamm0,,,tcr,,,,,);

				if rc<0 then
					do;
						call nlpnms(rc,pgammf,"critbisf",pgamm0,,,,,,);
					end;
			%end;
		%else
			%do;
				/* if no truncation is allowed, gamm is forced to 0 */
				pgammf=gamma_1(0);
			%end;

		cr=critbisf(pgammf);
		Cr1=Ct1;
		Cr2=Ct2;
		gammf=gammap(pgammf);

		* print gammf cr betaf n1t n2t;
		cri=cr;

		/* computation of the test statistics */
		start calctest;
			umin1=max(0,tg_1(0,gamm));
			umax1=min(1,tg_1(1,gamm));
			umin2=max(0,tg(0,gamm));
			umax2=min(1,tg(1,gamm));
			rang1=t((1:nrow(y1))/nrow(y1));
			rang2=t((1:nrow(y2))/nrow(y2));
			tg_rang2=tg_1(rang2,gamm);
			tgrang1=tg(rang1,gamm);
			l1=loc((umin1<=rang1)#(rang1<=umax1));
			l2=loc((umin2<=rang2)#(rang2<=umax2));
			yn1=y1[l1];
			yn2=y2[l2];
			n1t=nrow(yn1);
			n2t=nrow(yn2);

			* print n1t n2t;
			rangn1=rang1[l1];
			tg_rangn2=tg_rang2[l2];
			tgrangn1=tgrang1[l1];
			rangn2=rang2[l2];
			qy1=quant(y1,tg_rangn2);
			qy2=quant(y2,tgrangn1);
			y=((yn1||qy2||rangn1||tgrangn1)//(qy1||yn2||tg_rangn2||rangn2));

			/*calcul de mu1*/
			mu2=(zb_1(y[,2],beta)+y[,1])/2;
			mu1=(y[,2]+zb(y[,1],beta))/2;

			/* computation of quantile derivatives */
			dlambd1=BlGart(y[,3],y1);
			dlambd2=BlGart(y[,4],y2);
			dmu1=1/2*(dzb_u(y[,1],beta)#dlambd1*(umax1-umin1)+dtg_u(y[,4],gamm)#dlambd2*(umax2-umin2));
			eps=1e-8;
			correc=(umax2-umin2)/(umax1-umin1)/(dzb_u(y[,2],beta)##2)/(dtg_u(y[,4],gamm)+eps);
			vw1=1/(dmu1##2)/(1+correc);
			vw2=vw1;
			y=y||vw1||vw2||(shape(1,nrow(yn1),1)//shape(2,nrow(yn2),1));
			cr=critb(beta);
			stat=cr/(1/nrow(yn1)+1/nrow(yn2));
		finish;
		beta=betaf;
		gamm=gammap(pgammf);
		run calctest;

		stat1=stat;

		* print stat1;
		%if &shift=0 %then
			%do;
				name={'cr' 'gamm_e' 'D' 'Crit1' 'Crit2' 'n1t' 'n2t' 'stat1'};
				res=(cri || t(gamm) || betaf || Cr1 || Cr2 || n1t || n2t || stat1);
			%end;
		%else %if &dilation=0 %then
			%do;
				name={'cr' 'gamm_e' 'A' 'Crit1' 'Crit2' 'n1t' 'n2t' 'stat1'};
				res=(cri || t(gamm) || betaf || Cr1 || Cr2 || n1t || n2t || stat1);
			%end;
		%else
			%do;
				name={'cr' 'gamm_e' 'D' 'A' 'Crit1' 'Crit2' 'n1t' 'n2t' 'stat1'};
				res=(cri || t(gamm) || betaf[1] || betaf[2] || Cr1 || Cr2 || n1t || n2t || stat1);
			%end;

		create result from res [colname=name];
		append from res;
	quit;

	data result;
		set result;

		if abs(gamm_e)<1e-10 then
			gamm_e=0;
		&varcat.0=&varref0;
		&varcat.1=&varref1;
		truncation=&truncation;
		shift=&shift;
		dilation=&dilation;
	run;

%mend;

%macro estimt(variable,data,varcat,varref0,varref1,trim);

	proc sort data=&data(where=(&varcat in(&varref0,&varref1))) out=data_estimt(keep=&variable &varcat);
		by &varcat;
	run;

	/* threshold for the removing of extreme values */
	%let s1=&trim;
	%let s2=%eval(100-&trim.);

	proc univariate data=data_estimt noprint pctldef=3;
		by &varcat;
		var &variable;
		output out=thresholds pctlpre=prod pctlpts=&s1 &s2;
	run;

	data data_estimt;
		merge data_estimt thresholds;
		by &varcat;

		if prod&s1.<=&variable.<=prod&s2.;
	run;

	/* conditional mean of categ=0 put to 0, shift of every observations */
	proc summary data=data_estimt(where=(&varcat=&varref0));
		var &variable;
		output out=thresholds mean=mprod;
	run;

	data data_estimt;
		if _n_=1 then
			set thresholds(keep=mprod);
		set data_estimt;
		&variable=&variable-mprod;
	run;

	/* estimation without any transformation: computation of the basic criterium value */
	%estquant(&variable,data_estimt,&varcat,&varref0,&varref1,0,0,0);
	run;

	data resultf0;
		set result;
		criteria0=cr;
		obs=n1t+n2t;
		id=1;

		/* allowing for truncation, shift and dilation */
		%estquant(&variable,data_estimt,&varcat,&varref0,&varref1,1,1,1);
	run;

	data resultf;
		set resultf0 (drop=id criteria0 obs) result;

		/* allowing for truncation and shift */
		%estquant(&variable,data_estimt,&varcat,&varref0,&varref1,1,1,0);
	run;

	data resultf;
		set resultf result;

		/* allowing only for shift */
		%estquant(&variable,data_estimt,&varcat,&varref0,&varref1,0,1,0);
	run;

	data resultf;
		set resultf result;

		/* allowing only for truncation */
		%estquant(&variable,data_estimt,&varcat,&varref0,&varref1,1,0,0);
	run;

	data resultf;
		set resultf result;

		/* allowing for shift and dilation */
		%estquant(&variable,data_estimt,&varcat,&varref0,&varref1,0,1,1);
	run;

	data resultf;
		set resultf result;

		/* value of the truncation parameter */
		S=-gamm_e/(1-gamm_e);

		if A=. then
			A=0;

		if D=. then
			D=1;
		id=1;
		drop gamm_e cat0 cat1 stat1;
		rename cr=criteria;

		/* Computes the R2 */
	data resultf;
		merge resultf resultf0 (keep=id criteria0 obs);
		by id;
		R2=1-criteria/criteria0;
		drop id criteria0;

	data resultf;
		set resultf;
		attrib name length=$50;
		name="&name";
		id=_N_;

	data a;
		set resultf;
		keep name id;

	proc sort data=a;
		by id;

	proc sort data=resultf;
		by id;

	data a;
		merge a resultf (keep=shift dilation id);
		by id;

	data a;
		merge a resultf (keep=truncation A id);
		by id;

	data a;
		merge a resultf (keep=D S R2 id);
		by id;

	data a;
		merge a resultf (keep=obs id);
		by id;

	data resultff;
		merge a resultf (keep=criteria n1t n2t id);
		by id;
		drop id;
	run;

%mend estimt;
